/*
This module implements parsing of the STK500v2 protocol and assembly of
response messages. It calls functions in the isp and hvprog modules to
perform the real work.
*/

#pragma once

#include "utils.h"

namespace stk{

void IncrementAddress();

void SetRxChar(byte c);
int GetTxByte();
int GetTxCount();
void Poll();  /* must be called from main loop */

union StkParam{
 byte bytes[32];
 struct{
  int     buildVersionLow;
  byte   reserved1[14];
  byte   hardwareVersion;
  byte   softwareVersionMajor;
  byte   softwareVersionMinor;
  byte   reserved2;
  byte   vTarget;
  byte   vRef;
  byte   oscPrescale;
  byte   oscCmatch;
  byte   sckDuration;
  byte   reserved3;
  byte   topcardDetect;
  byte   reserved4;
  byte   status;
  byte   data;
  byte   resetPolarity;
  byte   controllerInit;
 };
};

/* ------------------------------------------------------------------------- */

extern Word<dword> Address;
extern StkParam stkParam;

/* ------------------------------------------------------------------------- */
/* ----------------------- command parameter structs ----------------------- */
/* ------------------------------------------------------------------------- */
namespace isp{

struct stkEnterProg{
 byte timeout;
 byte stabDelay;
 byte cmdExeDelay;
 byte synchLoops;
 byte byteDelay;
 byte pollValue;
 byte pollIndex;
 byte cmd[4];
};

struct stkLeaveProg{
 byte preDelay;
 byte postDelay;
};

struct stkChipErase{
 byte eraseDelay;
 byte pollMethod;
 byte cmd[4];
};

struct stkProgramFlash{
 byte numBytes[2];
 byte mode;
 byte delay;
 byte cmd[3];
 byte poll[2];
 byte data[256];
};

struct stkReadFlash{
 byte numBytes[2];
 byte cmd;
};

struct stkProgramFuse{
 byte cmd[4];
};

struct stkReadFuse{
 byte retAddr;
 byte cmd[4];
};

struct stkMulti{
 byte numTx;
 byte numRx;
 byte rxStartAddr;
 byte txData[256];
};

}/*namespace isp*/

/* ------------------------------------------------------------------------- */

namespace hv{

struct stkEnterProgSp{
 byte stabDelay;
 byte cmdExeDelay;
 byte synchCycles;
 byte latchCycles;
 byte toggleVtg;
 byte powerOffDelay;
 byte resetDelay1;
 byte resetDelay2;
};

struct stkLeaveProg{
 byte stabDelay;
 byte resetDelay;
};

struct stkChipEraseSp{
 byte pollTimeout;
 byte eraseTime;
};

struct stkReadFlash{
 byte numBytes[2];	// big-endian (ausgestorben aber knstlich am Leben erhalten von ewiggestrigen Dinosaurierfans)
};

struct stkProgramFlash:public stkReadFlash{
 byte mode;
 byte pollTimeout;
 byte data[1];    /* actually more data than 1 byte */
};

struct stkReadFuse{
 byte fuseAddress;
};

struct stkProgramFuseSp:public stkReadFuse{
 byte fuseByte;
 byte pollTimeout;
};

/* ------------------------------------------------------------------------- */

struct stkEnterProgPp{
 byte stabDelay;
 byte progModeDelay;
 byte latchCycles;
 byte toggleVtg;
 byte powerOffDelay;
 byte resetDelayMs;
 byte resetDelayUs;
};

struct stkChipErasePp{
 byte   pulseWidth;
 byte   pollTimeout;
};

struct stkProgramFusePp{
 byte address;
 byte data;
 byte pulseWidth;
 byte pollTimeout;
};
}/*namespace hv*/
/* ------------------------------------------------------------------------- */

#define STK_STX     27
#define STK_TOKEN   14

/* =================== [ STK general command constants ] =================== */

#define STK_CMD_SIGN_ON                         0x01
#define STK_CMD_SET_PARAMETER                   0x02
#define STK_CMD_GET_PARAMETER                   0x03
#define STK_CMD_SET_DEVICE_PARAMETERS           0x04
#define STK_CMD_OSCCAL                          0x05
#define STK_CMD_LOAD_ADDRESS                    0x06
#define STK_CMD_FIRMWARE_UPGRADE                0x07


/* =================== [ STK ISP command constants ] =================== */

#define STK_CMD_ENTER_PROGMODE_ISP              0x10
#define STK_CMD_LEAVE_PROGMODE_ISP              0x11
#define STK_CMD_CHIP_ERASE_ISP                  0x12
#define STK_CMD_PROGRAM_FLASH_ISP               0x13
#define STK_CMD_READ_FLASH_ISP                  0x14
#define STK_CMD_PROGRAM_EEPROM_ISP              0x15
#define STK_CMD_READ_EEPROM_ISP                 0x16
#define STK_CMD_PROGRAM_FUSE_ISP                0x17
#define STK_CMD_READ_FUSE_ISP                   0x18
#define STK_CMD_PROGRAM_LOCK_ISP                0x19
#define STK_CMD_READ_LOCK_ISP                   0x1A
#define STK_CMD_READ_SIGNATURE_ISP              0x1B
#define STK_CMD_READ_OSCCAL_ISP                 0x1C
#define STK_CMD_SPI_MULTI                       0x1D

/* =================== [ STK PP command constants ] =================== */

#define STK_CMD_ENTER_PROGMODE_PP               0x20
#define STK_CMD_LEAVE_PROGMODE_PP               0x21
#define STK_CMD_CHIP_ERASE_PP                   0x22
#define STK_CMD_PROGRAM_FLASH_PP                0x23
#define STK_CMD_READ_FLASH_PP                   0x24
#define STK_CMD_PROGRAM_EEPROM_PP               0x25
#define STK_CMD_READ_EEPROM_PP                  0x26
#define STK_CMD_PROGRAM_FUSE_PP                 0x27
#define STK_CMD_READ_FUSE_PP                    0x28
#define STK_CMD_PROGRAM_LOCK_PP                 0x29
#define STK_CMD_READ_LOCK_PP                    0x2A
#define STK_CMD_READ_SIGNATURE_PP               0x2B
#define STK_CMD_READ_OSCCAL_PP                  0x2C    

#define STK_CMD_SET_CONTROL_STACK               0x2D

/* =================== [ STK HVSP command constants ] =================== */

#define STK_CMD_ENTER_PROGMODE_HVSP             0x30
#define STK_CMD_LEAVE_PROGMODE_HVSP             0x31
#define STK_CMD_CHIP_ERASE_HVSP                 0x32
#define STK_CMD_PROGRAM_FLASH_HVSP              0x33
#define STK_CMD_READ_FLASH_HVSP                 0x34
#define STK_CMD_PROGRAM_EEPROM_HVSP             0x35
#define STK_CMD_READ_EEPROM_HVSP                0x36
#define STK_CMD_PROGRAM_FUSE_HVSP               0x37
#define STK_CMD_READ_FUSE_HVSP                  0x38
#define STK_CMD_PROGRAM_LOCK_HVSP               0x39
#define STK_CMD_READ_LOCK_HVSP                  0x3A
#define STK_CMD_READ_SIGNATURE_HVSP             0x3B
#define STK_CMD_READ_OSCCAL_HVSP                0x3C

/* =================== [ STK status constants ] =================== */

/* Success */
#define STK_STATUS_CMD_OK                       0x00

/* Warnings */
#define STK_STATUS_CMD_TOUT                     0x80
#define STK_STATUS_RDY_BSY_TOUT                 0x81
#define STK_STATUS_SET_PARAM_MISSING            0x82

/* Errors */
#define STK_STATUS_CMD_FAILED                   0xC0
#define STK_STATUS_CKSUM_ERROR                  0xC1
#define STK_STATUS_CMD_UNKNOWN                  0xC9

/* =================== [ STK parameter constants ] =================== */
#define STK_PARAM_BUILD_NUMBER_LOW              0x80
#define STK_PARAM_BUILD_NUMBER_HIGH             0x81
#define STK_PARAM_HW_VER                        0x90
#define STK_PARAM_SW_MAJOR                      0x91
#define STK_PARAM_SW_MINOR                      0x92
#define STK_PARAM_VTARGET                       0x94
#define STK_PARAM_VADJUST                       0x95
#define STK_PARAM_OSC_PSCALE                    0x96
#define STK_PARAM_OSC_CMATCH                    0x97
#define STK_PARAM_SCK_DURATION                  0x98
#define STK_PARAM_TOPCARD_DETECT                0x9A
#define STK_PARAM_STATUS                        0x9C
#define STK_PARAM_DATA                          0x9D
#define STK_PARAM_RESET_POLARITY                0x9E
#define STK_PARAM_CONTROLLER_INIT               0x9F

/* =================== [ STK answer constants ] =================== */

#define STK_ANSWER_CKSUM_ERROR                  0xB0

struct SerialMessage{
 byte stx,seq;
 byte beLen[2];
 byte token;
};

struct Param:public SerialMessage{
 byte maincmd;
 union{
  struct{
   byte index,value;
  }parameter;
  Word<dword> beAddress;
  isp::stkEnterProg	ispEnterProg;
  isp::stkLeaveProg	ispLeaveProg;
  isp::stkChipErase	ispChipErase;
  isp::stkProgramFlash	ispProgramFlash;
  isp::stkReadFlash	ispReadFlash;
  isp::stkProgramFuse	ispProgramFuse;
  isp::stkReadFuse	ispReadFuse;
  isp::stkMulti		ispMulti;
  hv::stkEnterProgSp	hvEnterProgSp;
  hv::stkEnterProgPp	hvEnterProgPp;
  hv::stkLeaveProg	hvLeaveProg;
  hv::stkChipEraseSp	hvChipEraseSp;
  hv::stkChipErasePp	hvChipErasePp;
  hv::stkReadFlash	hvReadFlash;
  hv::stkProgramFlash	hvProgramFlash;
  hv::stkReadFuse	hvReadFuse;
  hv::stkProgramFuseSp	hvProgramFuseSp;
  hv::stkProgramFusePp	hvProgramFusePp;
 };
};

struct Result:public SerialMessage{
 byte maincmd, status;
 union{
  char text[256];
  byte data[256];
 };
};

#define BUFFER_SIZE     281 /* results in 275 bytes max body size */
extern byte rxBuffer[BUFFER_SIZE],txBuffer[BUFFER_SIZE];

#define gParam (*reinterpret_cast<const Param*>(rxBuffer))
#define gResult (*reinterpret_cast<Result*>(txBuffer))

}/*namespace stk*/
